import os
import sys
import subprocess

# Function to install missing dependencies
def install_requirements():
    required_packages = [
        "gradio",
        "torch",
        "transformers",
        "diffusers",
        "Pillow",
        "numpy",
        "huggingface_hub"
    ]
    
    for package in required_packages:
        try:
            __import__(package)
        except ImportError:
            print(f"Installing missing package: {package}")
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])

# Install required dependencies before running the script
install_requirements()

import gradio as gr
import torch
from janus.janusflow.models import MultiModalityCausalLM, VLChatProcessor
from PIL import Image
from diffusers import DiffusionPipeline
import numpy as np
import random
from huggingface_hub import snapshot_download

# Ensure GPU is available
if not torch.cuda.is_available():
    raise EnvironmentError("GPU is required for this script. Please run on a machine with a CUDA-enabled GPU.")

# Use GPU for computations
cuda_device = 'cuda'

# Optimize GPU memory usage
torch.backends.cuda.matmul.allow_tf32 = True
torch.backends.cudnn.benchmark = True
torch.backends.cudnn.deterministic = False
torch.cuda.empty_cache()

# Hugging Face model cache directory
cache_dir = os.path.expanduser("~/.cache/huggingface/hub")

# Function to check if a model exists locally
def model_exists(model_name):
    model_path = os.path.join(cache_dir, f"models--{model_name.replace('/', '--')}")
    return os.path.exists(model_path)

# Load JanusFlow Model
def load_janusflow():
    global vl_gpt, vl_chat_processor, tokenizer
    if 'vl_gpt' not in globals():
        print("🔹 Loading JanusFlow Model...")
        model_name = "deepseek-ai/JanusFlow-1.3B"
        if not model_exists(model_name):
            print(f"🔽 Downloading {model_name}...")
            snapshot_download(repo_id=model_name, cache_dir=cache_dir)
        vl_chat_processor = VLChatProcessor.from_pretrained(model_name)
        tokenizer = vl_chat_processor.tokenizer
        vl_gpt = MultiModalityCausalLM.from_pretrained(
            model_name, torch_dtype=torch.float16
        ).to(cuda_device).eval()
        print("✅ JanusFlow Model Loaded.")

# Load FLUX Model
def load_flux():
    global flux_pipeline
    if 'flux_pipeline' not in globals():
        print("🔹 Loading FLUX Model...")
        model_name = "black-forest-labs/FLUX.1-schnell"
        if not model_exists(model_name):
            print(f"🔽 Downloading {model_name}...")
            snapshot_download(repo_id=model_name, cache_dir=cache_dir)
        torch.cuda.empty_cache()
        torch.cuda.ipc_collect()
        flux_pipeline = DiffusionPipeline.from_pretrained(
            model_name,
            torch_dtype=torch.float16,
            device_map="balanced",
        )
        try:
            if hasattr(flux_pipeline, "enable_xformers_memory_efficient_attention"):
                flux_pipeline.enable_xformers_memory_efficient_attention()
        except Exception as e:
            print(f"Warning: Could not enable memory efficient attention: {e}")
        print("✅ FLUX Model Loaded.")

# Ensure Safe Seed Function is Defined
def safe_seed(seed):
    return int(seed) % (2**32)

# Resize Large Images for Multimodal Understanding
def resize_large_image(image):
    max_size = 1024
    if isinstance(image, Image.Image):
        if max(image.size) > max_size:
            image.thumbnail((max_size, max_size), Image.ANTIALIAS)
    return image

# Multimodal Understanding Function
@torch.inference_mode()
def multimodal_understanding(image, question, seed, top_p, temperature):
    load_janusflow()
    torch.cuda.empty_cache()

    image = resize_large_image(image)

    if isinstance(image, Image.Image):
        image = np.array(image)

    seed = safe_seed(seed)
    torch.manual_seed(seed)
    np.random.seed(seed)
    torch.cuda.manual_seed(seed)

    conversation = [
        {"role": "User", "content": f"<image_placeholder>\n{question}", "images": [image]},
        {"role": "Assistant", "content": ""},
    ]

    pil_images = [Image.fromarray(image)]
    prepare_inputs = vl_chat_processor(
        conversations=conversation, images=pil_images, force_batchify=True
    ).to(cuda_device, dtype=torch.float16)

    inputs_embeds = vl_gpt.prepare_inputs_embeds(**prepare_inputs)

    outputs = vl_gpt.language_model.generate(
        inputs_embeds=inputs_embeds,
        attention_mask=prepare_inputs.attention_mask,
        pad_token_id=tokenizer.eos_token_id,
        max_new_tokens=400,
        do_sample=False if temperature == 0 else True,
        temperature=temperature,
        top_p=top_p
    )

    raw_response = tokenizer.decode(outputs[0].cpu().tolist(), skip_special_tokens=True)
    return raw_response

# Generate Image Function
@torch.inference_mode()
def generate_image(prompt, seed, cfg_weight, steps, randomize):
    load_flux()
    torch.cuda.empty_cache()  # Try clearing cache before starting
    torch.cuda.ipc_collect()
    
    if randomize:
        seed = random.randint(0, 2**32 - 1)
    seed = safe_seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)

    with torch.amp.autocast(device_type="cuda", dtype=torch.float16):
        result = flux_pipeline(
            prompt=prompt,
            guidance_scale=cfg_weight,
            num_inference_steps=steps,
            width=1024,
            height=1024
        )
        image = result.images[0]

    image.save("generated_image.png")

    # **Force memory cleanup after generation**
    del result
    torch.cuda.empty_cache()
    torch.cuda.ipc_collect()

    return image

# Random Seed Button Function
def randomize_seed_button():
    return random.randint(0, 2**32 - 1)

# Gradio UI
with gr.Blocks(title="Janus & FLUX Model") as demo:
    with gr.Tabs():
        with gr.Tab("Multimodal Understanding"):
            gr.Markdown("## Multimodal Understanding")
            with gr.Row():
                image_input = gr.Image(label="Upload Image", height=512, width=512)
                with gr.Column():
                    question_input = gr.Textbox(label="Chat with Image", placeholder="Ask a question about the image...")
                    seed_input = gr.Number(label="Seed", value=42, precision=0)
                    top_p = gr.Slider(0, 1, value=0.2, label="Top-p")
                    temperature = gr.Slider(0, 1, value=0.2, label="Temperature")
            response_output = gr.Textbox(label="Response", lines=4)
            chat_button = gr.Button("💬 Chat")

            # Preserve ALL buttons exactly as you had them
            description_buttons = [
                ("👤 Person Only", "Only describe the person. Ignore all background, setting, and lighting."),
                ("📝 Brief", "Brief length description."),
                ("📘 Medium", "Medium length description."),
                ("📗 Detailed", "Detailed length description."),
                ("📚 Highly Detailed", "Highly detailed description with intricate details and unique aspects.")
            ]

            with gr.Row():
                for btn_text, query in description_buttons:
                    gr.Button(btn_text).click(
                        multimodal_understanding,
                        inputs=[image_input, gr.Textbox(value=query, visible=False), seed_input, top_p, temperature],
                        outputs=response_output
                    )

            chat_button.click(
                multimodal_understanding,
                inputs=[image_input, question_input, seed_input, top_p, temperature],
                outputs=response_output
            )

        with gr.Tab("Text-to-Image Generation"):
            gr.Markdown("## Text-to-Image Generation")
            cfg_weight_input = gr.Slider(1, 10, value=3.6, label="CFG Weight")
            step_input = gr.Slider(1, 50, value=30, label="Number of Inference Steps")
            prompt_input = gr.Textbox(label="Prompt", lines=3)
            seed_input_tti = gr.Number(label="Seed (Optional)", value=123456789, precision=0)

            with gr.Row():
                randomize_seed_btn = gr.Button("Randomize Seed")
                generate_button = gr.Button("Generate Images")

            image_output = gr.Image(label="Generated Image", height=512, width=512)

            randomize_seed_btn.click(fn=randomize_seed_button, outputs=seed_input_tti)

            generate_button.click(generate_image, inputs=[prompt_input, seed_input_tti, cfg_weight_input, step_input, gr.State(False)], outputs=image_output)

demo.launch(share=True)
